# -*- coding: utf-8 -*-

"""
pipes.api
~~~~~~~~~
Class definitions to create, validate and run jobs as pipelines.

Description:

`Job` is the basic unit of pipeline which does some work.
`Pipe` is a structure which helps run the above jobs one after the other or in
parallel. A pipe can be used to run jobs or other pipes.

Hence overtly complicated pipelines can be boiled down to the above two basic
blocks.
"""

import logging
import subprocess
from collections import OrderedDict
from itertools import chain
from threading import Thread
from uuid import uuid4

import crayons

from .jobs import BashJob, Job, JobInterface


class Pipe(object):
    """Class to define the pipeline structure."""

    def __init__(self, name):
        """Constructor.

        :param name: Name given to the pipe object for identification.
        :type name: str
        """
        self.name = 'Pipe({})'.format(name)
        self.job_map = OrderedDict()

    def add_jobs(self, jobs, run_in_parallel=False):
        """Method to add jobs to pipeline.

        :param jobs: List of jobs/pipes to run.
        :type jobs: list
        :param run_in_parallel: This flag when set to False(default) runs the
        list of jobs given one after another. This flag if set to True runs the
        jobs/pipes submitted in parallel threads.
        :type run_in_parallel: boolean
        """

        # Return if nothing to do.
        if not jobs:
            return

        # Validate the set of jobs given.
        Pipe._validate(jobs)

        # Add jobs to pipeline.
        if run_in_parallel:
            self._add_in_parallel(jobs)
        else:
            self._add_in_series(jobs)

    def run(self):
        """Method to run the pipeline."""
        logging.debug('run() method called on {}'.format(self.name))
        for jobset in self.job_map.values():
            job_threads = []
            # Create job threads
            for job in jobset:
                job_threads.append(Pipe._create_thread_for(job))
            # Start job threads
            for job in job_threads:
                job.start()
            # Job main thread to create flow
            for job in job_threads:
                job.join()

    def graph(self):
        """Method to print the structure of the pipeline."""
        logging.debug('graph() method called on {}'.format(self.name))
        self._pretty_print()
        print('')

    @staticmethod
    def _create_thread_for(job):
        """Method to create thread to run the job in.

        :param job: Pipeline job to run.
        :type job: Job or BashJob or Pipe
        """
        if isinstance(job, Job):
            return Thread(
                target=job.function,
                args=job.args,
                kwargs=job.kwargs)
        elif isinstance(job, BashJob):
            # Note that without lambda, subprocess.Popen runs immediately.
            return Thread(target=lambda: subprocess.Popen(job.cmd).wait())
        else:
            return Thread(target=job.run)

    @staticmethod
    def _validate(jobs):
        """Method to validate the jobs submitted to pipeline.

        :param jobs: List of jobs submitted.
        :type jobs: list
        """
        for job in jobs:
            valid = isinstance(job, JobInterface) or isinstance(job, Pipe)
            if not valid:
                logging.error('Pipeline jobs should be of type Job or' +
                              ' BashJob or Pipe')
                raise AssertionError(
                    'Invalid type {} submitted'.format(type(job)))

    def _add_in_parallel(self, jobs):
        """Method to add jobs to pipeline so that they run in parallel.

        :param jobs: List of jobs submitted.
        :type jobs: list
        """
        self.job_map[uuid4()] = jobs
        logging.debug('{} submitted to be run in parallel'.format(jobs))

    def _add_in_series(self, jobs):
        """Method to add jobs to pipeline so that they run one after another,
        only after the previous job completes.

        :param jobs: List of jobs submitted.
        :type jobs: list
        """
        for job in jobs:
            self.job_map[uuid4()] = [job]
            logging.debug('{} submitted to be run in series'.format(job))

    def _pretty_print(self):
        """Method to pretty print the pipeline."""
        print('')
        print(crayons.green(self.name, bold=True))

        if not self.job_map:
            print(crayons.red('No jobs added to the pipeline yet !'))
            return

        joblen = len(self.job_map)
        for i, jobs in enumerate(self.job_map.values()):
            print(crayons.blue(u'| '))
            if len(jobs) == 1:
                print(crayons.blue(u'\u21E8  ') + crayons.white(jobs[0].name))
            else:
                if i == joblen - 1:
                    pre = u'  '
                else:
                    pre = u'| '
                l1 = [u'-' * (len(j.name) + 2) for j in jobs]
                l1 = u''.join(l1)
                l1 = l1[:-len(jobs[-1].name) // 2 + 1]
                print(crayons.blue(u'\u21E8 ') + crayons.blue(l1))
                fmt = u'{0:^{wid}}'
                l2 = [fmt.format(u'\u21E9', wid=len(j.name) + 2) for j in jobs]
                print(crayons.blue(pre) + crayons.blue(u''.join(l2)))
                l3 = [fmt.format(j.name, wid=len(j.name) + 2) for j in jobs]
                print(crayons.blue(pre) + crayons.white(u''.join(l3)))

        pipes = filter(lambda x: isinstance(x, Pipe),
                       chain(*self.job_map.values()))

        for item in pipes:
            item._pretty_print()

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.__repr__()



# -*- coding: utf-8 -*-

"""
jobs.api
~~~~~~~~
Class definitions to describe python method calls in a deterministic way.

Description: Method outputs are determined by the inputs provided to it, hence
to describe all aspects of the method call we need to know what method is being
called and what arguments it is called with. Classes in this file help to
describe python methods in a structured way.
"""

from inspect import isroutine


class JobInterface(object):
    """Pipeline job interface."""

    def __init__(self, name):
        """Constructor.

        :param name: Name of job.
        :type name: str
        """
        self.name = name


class Job(JobInterface):
    """Class to format python methods as pipeline jobs."""

    def __init__(self, function, args=(), kwargs={}):
        """Constructor.

        :param function: Python method to run.
        :type function: function
        :param args: Argument list to run the method with.
        :type args: tuple
        :param kwargs: Keyword arguments to run the method with.
        :type kwargs: dict
        """

        # Validate the structure of the job submitted.
        assert isroutine(function), 'Python callable expected'
        assert isinstance(args, tuple)
        assert isinstance(kwargs, dict)

        super(Job, self).__init__(function.__name__)
        self.function = function
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        return 'Job(function={}, args={}, kwargs={})'.format(
            self.name,
            self.args,
            self.kwargs)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        # Note that same method run with two different sets of parameters is
        # considered to be two different jobs and not one job.
        return self.function == other.function and \
            self.args == other.args and \
            self.kwargs == other.kwargs


class BashJob(JobInterface):
    """Class to format bash commands as pipeline jobs."""

    def __init__(self, cmd):
        """Constructor.

        :param cmd: Bash command to run.
        :type cmd: list
        """

        # Validate the structure of the job submitted.
        assert isinstance(cmd, list), 'Bash command as list of strings needed'

        super(BashJob, self).__init__(' '.join(cmd))
        self.cmd = cmd

    def __repr__(self):
        return 'BashJob(cmd={})'.format(self.name)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return self.cmd == other.cmd



#!/usr/bin/env python
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file.
with open(path.join(here, 'README.rst')) as f:
    long_description = f.read()

# Get package and author details.
about = {}
with open(path.join(here, 'pypette', '__version__.py')) as f:
    exec(f.read(), about)

setup(
    # Name of the module
    name=about['__title__'],

    # Details
    version=about['__version__'],
    description=about['__description__'],
    long_description=long_description,

    # The project's main homepage.
    url=about['__url__'],

    # Author details
    author=about['__author__'],
    author_email=about['__author_email__'],

    # License
    license=about['__license__'],
    packages=[about['__title__']],
    test_suite='tests',
    keywords='pipeline dev-tool threads multithreading python',
    classifiers=[
        # Intended Audience.
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        # License.
        'License :: OSI Approved :: MIT License',
        # Project maturity.
        'Development Status :: 3 - Alpha',
        # Operating Systems.
        'Operating System :: POSIX',
        # Supported Languages.
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        # Topic tags.
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=['crayons'])
